﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ZMapReader
{
    //simple class for ready ZMap data files commonly used in the oil gas field for geology related data (depth, temperature, etc).
    //Zmaps describe a rectangular grid and the data for the cells of the grid.  See comments in readHeader for structure.
    public class ZmapReader
    {
        private bool _hasreadheader = false;
        private String _path;
        public String path
        {
            get
            {
                return _path;
            }
            set
            {
                _path = value;
            }
        }
        private int _fieldwidth;
        private double _nullvalue;
        private string _nullvaluestr;
        private int _gridNodesPerLine;
        private StreamReader _streamReader;

        public int _y_ext;
        public int _x_ext;
        public double _x_min;
        public double _x_max;
        public double _y_min;
        public double _y_max;
        public double _dx;
        public double _dy;
        private float[] _data;

        public void readHeader()
        {
            _streamReader = new StreamReader(path);
            try
            {
                String line = _streamReader.ReadLine();
                while (!line.StartsWith("@"))//skip past any comments
                {
                    line = _streamReader.ReadLine();
                }
                //first non comment or empty line in header
                //On the same line as the first "@" there are three fields.  The first is user defined but many times is just "GRID FILE",
                //the second, for a grid file, must be "GRID", the third is an integer that indicates the number of grid nodes per physical line.           
                string[] fields = line.Split(',');
                String gstr = fields[1].Trim();
                if (gstr.CompareTo("GRID") != 0)
                {
                    throw new Exception("Expected GRID in zmap file");
                }
                gstr = fields[2].Trim();
                _gridNodesPerLine = Int32.Parse(gstr);
                //After the file description line are three more header lines that must exist for a grid file.
                //The first of these three lines has five fields.  The first field is the field width of each grid node as stored in the data
                //section below the last "@".  The second field is the missing or null data value as it will be found in the data section.  
                //The third field is a user defined text value used to indicate a missing or null value.  The text value is used if field two is blank.
                //If field two is defined, then the third field is left blank.  The forth field indicates the number of decimal places to use if
                //no decimal point is found in the data nodes.  (Historical note: when disk and tape space was at a premium, this field made it possible 
                //to save some space in the file by not having to store a "." for each data node.)  If the data values are written to the file with decimals,
                //this value is ignored.  The fifth field indicates the starting column of the first grid node on each line in the data section of the file.
                line = _streamReader.ReadLine();
                //we don't really care about any of those settings except for the null value which we will then later filter out all points that have any null values
                fields = line.Split(',');
                if (fields.Count() != 5)
                {
                    throw new Exception("Expected 5 fields in zmap file");
                }
                gstr = fields[0].Trim();
                _fieldwidth = Int32.Parse(gstr);
                gstr = fields[1].Trim();
                _nullvaluestr = gstr;
                _nullvalue = Double.Parse(gstr);

                //The second of the three lines has six fields.  The first field is the number of rows in the grid.  The second field is the number of
                //columns in the grid.  The third is the minimum grid X node value.  The forth is the maximum grid X node value.  The fifth is the minimum
                //grid Y node value.  The sixth is the maximum grid Y node value.  The last four fields define the bounding box of the grid, and implies,
                //for example, the lower left corner of the of the grid is the lower left corner of the lower left node in the grid.  The projection and 
                //or datum of the bounding box value is not defined.
                line = _streamReader.ReadLine();
                fields = line.Split(',');
                if (fields.Count() != 6)
                {
                    throw new Exception("Expected 6 fields in zmap file");
                }
                gstr = fields[0].Trim();
                _y_ext = Int32.Parse(gstr);
                gstr = fields[1].Trim();
                _x_ext = Int32.Parse(gstr);
                gstr = fields[2].Trim();
                _x_min = Double.Parse(gstr);
                gstr = fields[3].Trim();
                _x_max = Double.Parse(gstr);
                gstr = fields[4].Trim();
                _y_min = Double.Parse(gstr);
                gstr = fields[5].Trim();
                _y_max = Double.Parse(gstr);

                // Set cell size in X and Y.
                _dx = (_x_max - _x_min) / (_x_ext - 1);
                _dy = (_y_max - _y_min) / (_y_ext - 1);

                //The third of the three lines is always "0,0, 0.0, 0.0".
                line = _streamReader.ReadLine();
                //After the last header line, there is a single line with a single "@", the line after which is the first line of data in the data section
                //of the file.  The data section has fixed field widths each field being a single grid node, and is generally right justified.  
                //There will be no more nodes on any physical line than that defined in third field of the first header line.  A data field may or may not
                //have a decimal point.  If none is found, it is implied, and the decimal places are as defined in the second header line in the forth field.
                while (!line.StartsWith("@"))
                {
                    line = _streamReader.ReadLine();
                }
                //The grid nodes in the data section are stored in column major order.  That is the first column of data is written first, starting at the 
                //upper left corner of the grid.  For example, if the grid has 7 rows and three columns, and the number of nodes per line is 4,
                //the first line of the data section will have 4 nodes, the first four grid nodes going down from the upper left.  The second line will
                //have three nodes, the last three nodes of the first column.  Then the next column is written, four nodes then three.  Then the last 
                //column is written in the same pattern.
                //Here is an example of a 6 row x 4 column ZMapPlus grid file:


                //!
                //! File created by DMBTools2.GridFileFormats.ZmapPlusFile
                //!
                //@GRID FILE, GRID, 4
                //20, -9999.0000000, , 7, 1
                //6, 4, 0, 200, 0, 300
                //0.0, 0.0, 0.0
                //@
                //       -9999.0000000       -9999.0000000           3.0000000          32.0000000
                //          88.0000000          13.0000000
                //       -9999.0000000          20.0000000           8.0000000          42.0000000
                //          75.0000000           5.0000000
                //           5.0000000         100.0000000          35.0000000          50.0000000
                //          27.0000000           1.0000000
                //           2.0000000          36.0000000          10.0000000           6.0000000
                //           9.0000000       -9999.0000000

                //above comments on zmap format taken from http://www.osgeo.org/pipermail/gdal-dev/2011-June/029173.html by David Baker (Geoscience) david.m.baker at chk.com
                _hasreadheader = true;
            }
            catch (Exception ex)
            {
                if (_streamReader != null)
                {
                    _streamReader.Close();
                }
                throw ex;
            }
        }

        public float[] Data
        {
            get
            {
                return _data;
            }
        }

        public Double NullValue
        {
            get
            {
                return _nullvalue;
            }
            set
            {
                _nullvalue = value;
            }
        }
        private List<string> Split(string value, int chunkSize)
        {
            List<string> triplets = new List<string>();
            for (int i = 0; i < value.Length; i += chunkSize)
                if (i + chunkSize > value.Length)
                    triplets.Add(value.Substring(i));
                else
                    triplets.Add(value.Substring(i, chunkSize));

            return triplets;
        }
        public void ReadData()
        {
            if (!_hasreadheader)
            {
                readHeader();
            }
            try
            {
                _data = new float[(int)_x_ext * (int)_y_ext];
                int numlinestoread = (int)Math.Ceiling((double)_y_ext / _gridNodesPerLine);
                //char[] split = new char[] { ' ', '\n', '\r' };
                int index = 0;
                string line;
                while ((line = _streamReader.ReadLine()) != null)
                {
                    StringBuilder strbuf = new StringBuilder(line);
                    int i = 0;
                    while (i < numlinestoread - 1 &&
                           (line = _streamReader.ReadLine()) != null)
                    {
                        strbuf.Append(line);
                        i++;
                    }
                    line = strbuf.ToString();               
                    line = line.Replace("\n", string.Empty);                   
                    List<string> datastr = Split(line, _fieldwidth);
                    if (datastr.Count() != _y_ext)
                    {
                        throw new Exception("Invalid zmap file, not enough entries");
                    }
                    foreach (string dstr in datastr)
                    {
                        float f = Single.Parse(dstr);
                        _data[index] = ((int)f == (int)_nullvalue) ? Single.NaN : f;//nullvalues should always be ints
                        index++;
                    }
                }
            }
            finally
            {
                _streamReader.Close();
            }
        }

    }
}
